from unittest import TestCase
import numpy as np
from osirisl1services.services import (get_orbit_number, _get_odin_position_or_velocity_eci, get_odin_position_eci,
                                       Level1Services, get_scannos_in_orbit_below_sza, SPECTROGRAPH_ROTATION_TB_ZYX,
                                       IR_ROTATION_TB_ZYX)
from osirisl1services.readlevel1 import Orbit, open_level1_attitude, open_level1_spectrograph, open_level1_ir
import osl1 as l1
from osirisl1services.spatialrotation import ECI, ECEF, Geodetic
import sasktran as sk
from osirisl1services.psf_fit import PointSpreadFunctionFit


class TestGetOrbitNumber(TestCase):
    """ Tests all possible branches of the get_orbit_number method. """
    def found_here(self):
        self.assertEqual(52832, get_orbit_number(55500))

    def not_here_found_previous(self):
        self.assertEqual(40723, get_orbit_number(54690.0013))

    def not_here_found_next(self):
        self.assertEqual(52832, get_orbit_number(55499.9822))

    def not_here_not_previous(self):
        with self.assertRaises(LookupError):
            get_orbit_number(52205.199817080356)

    def not_here_not_next(self):
        with self.assertRaises(LookupError):
            get_orbit_number(52204.930533753504)

    def except_here_found_previous(self):
        """ Tests both branches where orbit data is missing for the closest orbit, but the previous orbit has attitude
            data. """
        self.assertEqual(6006, get_orbit_number(52364.81016006676))
        with self.assertRaises(LookupError):
            get_orbit_number(52204.930533953506)

    def except_here_found_next(self):
        """ Tests both branches where orbit data is missing for the closest orbit, but the next orbit has attitude
            data. """
        self.assertEqual(5682, get_orbit_number(52342.96757868986))
        with self.assertRaises(LookupError):
            get_orbit_number(52205.199816880355)

    def except_here_except_previous(self):
        with self.assertRaises(LookupError):
            get_orbit_number(58299.97)

    def except_here_except_next(self):
        with self.assertRaises(LookupError):
            get_orbit_number(58300)


class TestLevel1Services(TestCase):
    def setUp(self):
        self.scannos = [6432013,  # catches #25
                        27110030,  # catches #25
                        5626000]  # catches #60

    def test_get_odin_position_or_velocity_eci_missing_kind(self):
        with self.assertRaises(KeyError):
            _get_odin_position_or_velocity_eci(np.arange(55000, 55000.1, 0.01), kind=None)

    def test_get_odin_position_eci_for_mjd_list(self):
        actual = np.array(
            [[3138482.02769903, -333123.95920782, -6213003.2823842],
             [-3066070.16850131, -1347894.42544017, -6111605.41709904],
             [-6770576.5475889, -1265013.45938622, -1024132.20494053],
             [-4929966.08488909, -146780.56610887, 4899424.01108996],
             [955764.57233386, 1092033.03203767, 6793673.74919451],
             [6057015.62170004, 1435188.52236848, 3102498.16800911],
             [6187520.17890978, 600504.23510872, -3141816.78955488],
             [1256761.78266777, -726342.95457509, -6816668.84867819],
             [-4697849.4220833, -1462885.89773127, -4934141.66823197],
             [-6817047.37847959, -1006335.04223945, 978118.79582409]])
        np.testing.assert_allclose(get_odin_position_eci(np.arange(55000, 55000.1, 0.01)), actual, atol=1e-9)

    def test_get_odin_position_eci_for_mjd_list_missing_attitude_data(self):
        with self.assertRaises(LookupError):
            get_odin_position_eci(np.arange(52216.02, 52216.15, 0.001))

    def test_odin_velocity_ecef(self):
        """ Test that odin's velocity from c++ Level1 Services is the same as derived by the services
            coordinate transform code.
        """
        for scanno in self.scannos:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            exposure_mjd = spectrograph.l1._exposuremjd
            velocity = spectrograph.l1.velocity_ecef
            for mjd, vel in zip(exposure_mjd.values, velocity.values):
                velocity_eci = ECI(*l1.osl1.GetOdinVelocity(mjd)[1])
                velocity_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(velocity_eci))[1])
                self.assertAlmostEqual(velocity_ecef.x, vel[0], delta=0.5, msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(velocity_ecef.y, vel[1], delta=0.5, msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(velocity_ecef.z, vel[2], delta=0.5, msg=f'ScanNo: {scanno}')

    def test_spectrograph_look_cf(self):
        """ Test that the look direction from c++ Level1 Services is the same as derived by the TB angles in
            /OsirisConfigurationDatabase/flightduration/osoptics.ini

            Vector(x=0.99999495087295287, y=-0.0027594116765072515, z=0.0015760316621056657)
        """
        osiris_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.OS_X_BORE)[1]
        q = SPECTROGRAPH_ROTATION_TB_ZYX.to_quaternion()
        np.testing.assert_allclose(q.rot_unit_x(), osiris_look_cf)

    def test_pointing_for_spectrograph_dataset(self):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        for scanno in self.scannos + [47104027]:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            scan_info_swig, _, _, _, _ = l1.LoadScanOS(scanno)
            for (info, look_eci, look_ecef) in zip(scan_info_swig,
                                                   spectrograph.l1.look_eci.values,
                                                   spectrograph.l1.look_ecef.values):
                mjd = info.mjd + 0.5 * info.exposureTime / 86400.0
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1]
                np.testing.assert_allclose(look_eci, osiris_look_eci, atol=2e-7,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)
                osiris_look_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(osiris_look_eci))[1])
                np.testing.assert_allclose(look_ecef, osiris_look_ecef, atol=2e-7,
                                           err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_level1_spectrograph_SSC_tangent_point(self):
        """ Test comparing the tangent point derived by the attitude file to the tangent point returned
            by the c++ Level1 Services.
        """
        for scanno in self.scannos:
            attitude = open_level1_attitude(Orbit(scanno // 1000), scanno=scanno, load_pointing=True)
            for mjd, att in attitude.groupby('mjd'):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                position_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(position_eci))[1])
                osiris_look_cf = l1.osl1.GetCFUnitVector(mjd, l1.osl1.OS_X_BORE)[1]
                osiris_look_eci = ECI(*l1.osl1.CFToECI(mjd, osiris_look_cf, l1.osl1.ECI_SPACECRAFT)[1])
                osiris_look_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(osiris_look_eci))[1])
                tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci), list(osiris_look_eci), False)[1])
                tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])

                tangent_point_sasktran = sk.LineOfSight(mjd=mjd,
                                                        observer=position_ecef,
                                                        look_vector=osiris_look_ecef).tangent_location()

                self.assertAlmostEqual(tangentpoint.latitude, tangent_point_sasktran.latitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.longitude, tangent_point_sasktran.longitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')
                self.assertAlmostEqual(tangentpoint.height * 1000.0, tangent_point_sasktran.altitude, delta=1e-6,
                                       msg=f'ScanNo: {scanno}')

    def test_get_tangent_point_data_for_spectrograph_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the spectrograph.
        """
        for scanno in self.scannos:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            _, _, tangentpoint, _, _ = l1.LoadScanOS(scanno)
            np.testing.assert_allclose(spectrograph.l1.latitude, tangentpoint[0], atol=0.01,
                                       err_msg=f'ScanNo: {scanno}')
            np.testing.assert_allclose(spectrograph.l1.longitude, tangentpoint[1], atol=0.01,
                                       err_msg=f'ScanNo: {scanno}')
            np.testing.assert_allclose(spectrograph.l1.altitude, tangentpoint[2] * 1000.0, atol=10.0,
                                       err_msg=f'ScanNo: {scanno}')

    def test_get_tangent_point_data_for_spectrograph_dataset_contains_negatives(self):
        """ Test that certain position / look vectors return negative tangent points. """
        scanno = 51236096
        spectrograph = open_level1_spectrograph(scanno=scanno)
        self.assertTrue((spectrograph.l1.altitude < 0.0).any(),
                        msg=f'NaN tangent points were not found in ScanNo: {scanno}.')

    def test_solar_angles_for_spectrograph_dataset(self):
        """ Test comparing the solar angle solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the spectrograph.
        """
        for scanno in self.scannos:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            _, _, _, solar, _ = l1.LoadScanOS(scanno)
            np.testing.assert_allclose(spectrograph.l1.sza, solar[0], atol=0.01,
                                       err_msg=f'SZA mismatch. ScanNo: {scanno}')
            np.testing.assert_allclose(spectrograph.l1.ssa, solar[2], atol=0.01,
                                       err_msg=f'SSA mismatch. ScanNo: {scanno}')
            np.testing.assert_allclose(spectrograph.l1.saa, solar[1], atol=0.01,
                                       err_msg=f'SAA mismatch. ScanNo: {scanno}')

    def test_satellite_azimuth_angle_for_spectrograph_dataset(self):
        """ Test comparing the satellite look azimuth angle solution based on the attitude and look data with a
            precomputed value for the spectrograph.
        """
        for scanno in [6432013]:
            spectrograph = open_level1_spectrograph(scanno=scanno)
            sat_aa = [245.930041, 243.14435, 240.828005, 238.69051, 237.145373, 235.749343, 234.39029, 233.201785,
                      232.087428, 231.022396, 229.987813, 229.022138, 228.08295, 226.927345, 226.064481, 225.227268,
                      224.429019, 223.636542, 222.851146, 222.073715, 221.30497, 220.588469, 219.907575, 219.269356]
            np.testing.assert_allclose(spectrograph.l1.satellite_azimuth_angle, sat_aa, atol=0.01,
                                       err_msg=f'Satellite azimuth angle mismatch. ScanNo: {scanno}')

    def test_ir_look_cf(self):
        """ Test that the look direction from c++ Level1 Services is the same as derived by the TB angles in
            /OsirisConfigurationDatabase/flightduration/iroptics.ini

            Vector(x=0.999999420741261, y=0.0, z=0.0010763443418775663)
        """
        ir1_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR1_X_BORE)[1]
        ir2_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR2_X_BORE)[1]
        ir3_look_cf = l1.osl1.GetCFUnitVector(55000.0, l1.osl1.IR3_X_BORE)[1]
        q = IR_ROTATION_TB_ZYX.to_quaternion()
        np.testing.assert_allclose(q.rot_unit_x(), ir1_look_cf)
        np.testing.assert_allclose(q.rot_unit_x(), ir2_look_cf)
        np.testing.assert_allclose(q.rot_unit_x(), ir3_look_cf)

    def pointing_for_ir_channel_dataset(self, channel: int):
        """ Test comparing the look vector solution based on the attitude quaternion data to the look vectors returned
            by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scannos:
            ir = open_level1_ir(Orbit(scanno // 1000), channel=channel, load_audit=True)
            # For the purposes of testing, limit to the first 3 mjds
            ir = ir.isel(mjd=slice(3))
            for (mjd, look_eci, look_ecef) in zip(ir.l1._exposuremjd.values,
                                                  ir.l1.look_eci.values,
                                                  ir.l1.look_ecef.values):
                for pixel in range(128):
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    np.testing.assert_allclose(look_eci[pixel], pixel_look_eci, atol=1e-7,
                                               err_msg=f'ScanNo: {scanno}', verbose=True)
                    pixel_look_ecef = ECEF(*l1.osl1.ECIToGEO(mjd, list(pixel_look_eci))[1])
                    np.testing.assert_allclose(look_ecef[pixel], pixel_look_ecef, atol=1e-7,
                                               err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_pointing_for_ir_channel1_dataset(self):
        """ Test comparing the IR channel 1 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=1)

    def test_pointing_for_ir_channel2_dataset(self):
        """ Test comparing the IR channel 2 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=2)

    def test_pointing_for_ir_channel3_dataset(self):
        """ Test comparing the IR channel 3 look vector solution based on the attitude quaternion data to the look
            vectors returned by the c++ Level1 Services on points between the attitude solution for a dataset.
        """
        self.pointing_for_ir_channel_dataset(channel=3)

    def tangent_point_for_ir_dataset(self, channel: int):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR imager.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scannos:
            ir = open_level1_ir(Orbit(scanno // 1000), channel=channel)
            # For the purposes of testing, process a subset of an orbit
            ir = ir.isel(mjd=slice(0, -1, 100))
            for (mjd, lat, lon, alt) in zip(ir.l1._exposuremjd.values, ir.l1.latitude.values,
                                            ir.l1.longitude.values, ir.l1.altitude.values):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                # Assert the ECI positions are the to 5 meters
                for pixel in range(128):  # Test a subset of pixels for speed.
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci),
                                                                    list(pixel_look_eci), False)[1])
                    tangentpoint = Geodetic(*l1.osl1.ECIToGeodetic(mjd, list(tangentpoint_eci))[1])
                    if np.isfinite(float(lat[pixel])):  # Do not compare NaNs
                        self.assertAlmostEqual(float(lat[pixel]), tangentpoint.latitude, delta=0.01,
                                               msg=f'ScanNo: {scanno}')
                        self.assertAlmostEqual(float(lon[pixel]), tangentpoint.longitude, delta=0.01,
                                               msg=f'ScanNo: {scanno}')
                        self.assertAlmostEqual(float(alt[pixel]), tangentpoint.height * 1000.0, delta=10,
                                               msg=f'ScanNo: {scanno}')

    def test_tangent_point_for_ir_channel1_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 1.
        """
        self.tangent_point_for_ir_dataset(channel=1)

    def test_tangent_point_for_ir_channel2_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 2.
        """
        self.tangent_point_for_ir_dataset(channel=2)

    def test_tangent_point_for_ir_channel3_dataset(self):
        """ Test comparing the tangent point solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 3.
        """
        self.tangent_point_for_ir_dataset(channel=3)

    def solar_angles_for_ir_dataset(self, channel: int):
        """ Test comparing the solar angle solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR imager.
        """
        l1_inst_dict = {1: l1.osl1.INST_IR1, 2: l1.osl1.INST_IR2, 3: l1.osl1.INST_IR3}
        for scanno in self.scannos:
            ir = open_level1_ir(Orbit(scanno // 1000), channel=channel)
            # For the purposes of testing, process a subset of an orbit
            ir = ir.isel(mjd=slice(3))
            for (mjd, sza, ssa, saa) in zip(ir.l1._exposuremjd.values,
                                            ir.l1.sza.values, ir.l1.ssa.values, ir.l1.saa.values):
                position_eci = ECI(*l1.osl1.GetOdinPosition(mjd)[1])
                for pixel in [0, 1, 63, 64, 126, 127]:  # Test a subset of pixels for speed.
                    pixel_look = l1.osl1.GetPixelCFUnitVectorExt(mjd, l1_inst_dict[channel], 0.0, pixel + 0.5)[1]
                    pixel_look_eci = ECI(*l1.osl1.CFToECI(mjd, pixel_look, l1.osl1.ECI_SPACECRAFT)[1])
                    tangentpoint_eci = ECI(*l1.osl1.LOSTangentPoint(list(position_eci),
                                                                    list(pixel_look_eci), False)[1])
                    solar = l1.osl1.GetSolarAngles(mjd, tangentpoint_eci, pixel_look_eci)
                    if np.isfinite(float(sza[pixel])):  # Do not compare NaNs
                        self.assertAlmostEqual(float(sza[pixel]), solar[1], delta=0.01,
                                               msg=f'SZA mismatch. ScanNo: {scanno}')
                    if np.isfinite(float(ssa[pixel])):  # Do not compare NaNs
                        self.assertAlmostEqual(float(ssa[pixel]), solar[3], delta=0.01,
                                               msg=f'SSA mismatch. ScanNo: {scanno}')
                    if np.isfinite(float(saa[pixel])):  # Do not compare NaNs
                        self.assertAlmostEqual(float(saa[pixel]), solar[2], delta=0.01,
                                               msg=f'SAA mismatch. ScanNo: {scanno}')

    def test_satellite_azimuth_angle_for_ir_dataset(self):
        """ Test comparing the satellite look azimuth angle solution based on the attitude and look data with a
            precomputed value for the IR imager.
        """
        orbit = 6432
        channel = 1
        ir = open_level1_ir(Orbit(orbit), channel=channel)
        ir = ir.isel(mjd=[0, 1450])
        # Near the equator, the satellite azimuth angle should be around -8 for ascending and 188 for descending
        sat_aa = [351.566935, 188.533204]
        np.testing.assert_allclose(ir.l1.satellite_azimuth_angle.sel(pixel=0), sat_aa,
                                   err_msg=f'Satellite azimuth angle mismatch. Orbit: {orbit}')

    def test_solar_angles_for_ir_channel1_dataset(self):
        """ Test comparing the solar angle solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 1.
        """
        self.solar_angles_for_ir_dataset(channel=1)

    def test_solar_angles_for_ir_channel2_dataset(self):
        """ Test comparing the solar angle solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 2.
        """
        self.solar_angles_for_ir_dataset(channel=2)

    def test_solar_angles_for_ir_channel3_dataset(self):
        """ Test comparing the solar angle solution based on the attitude and look data with the c++ Level1 Services
            on points between the attitude solution for the IR channel 3.
        """
        self.solar_angles_for_ir_dataset(channel=3)

    def test_scans_in_orbit(self):
        """ Test the service returning scans in an orbit.
        """
        self.assertListEqual([scanno for scanno in get_scannos_in_orbit_below_sza(Orbit(6432), sza=90.0)],
                             [6432000, 6432001, 6432002, 6432003, 6432004, 6432005, 6432006, 6432007, 6432008, 6432009,
                              6432010, 6432011, 6432012, 6432013, 6432014, 6432015, 6432016, 6432017, 6432018, 6432019,
                              6432020, 6432021, 6432022, 6432023, 6432024, 6432025, 6432026, 6432027, 6432028, 6432063,
                              6432064, 6432065])

    def test_no_scans_in_orbit_with_attitude_file(self):
        """ Test the service returning scans in an orbit raises StopIteration when there are no scans.
        """
        with self.assertRaises(StopIteration):
            next(get_scannos_in_orbit_below_sza(Orbit(6000), sza=90.0))

    def test_no_scans_in_orbit_no_attitude_file(self):
        """ Test the service returning scans with a scan number without an attitude file
        """
        with self.assertRaises(StopIteration):
            next(get_scannos_in_orbit_below_sza(Orbit(6462), sza=90.0))

    def test_psf_fit_residuals(self):
        """ Test that output of PointSpreadFunctionFit.fit() in PythonLevel1Services matches output of the
            stand-alone project.
        """
        psf_test_dict = {6432013: {'fwhm': 0.9262741999408295,
                                   'shift': 0.05446324297262721},
                         27110030: {'fwhm': 0.9395053846503276,
                                    'shift': 0.023467222524174592},
                         4097029: {'fwhm': 0.932808490872749,
                                   'shift': 0.06667136518077217}}

        for scanno in psf_test_dict:
            psf_results = PointSpreadFunctionFit(scanno).fit()
            self.assertAlmostEqual(psf_test_dict[scanno]['fwhm'], psf_results.fwhm, places=6)
            self.assertAlmostEqual(psf_test_dict[scanno]['shift'], psf_results.shift, places=6)

    def test_spectrograph_load_psf(self):
        """ Test that the PSF corrections are applied. """
        for scanno in self.scannos:
            unapplied = open_level1_spectrograph(scanno=scanno)
            applied = open_level1_spectrograph(scanno=scanno, load_psf=True)
            # Pixels 575 and up should be the same for both wavelength and fwhm
            np.testing.assert_allclose(unapplied.wavelength[575:], applied.wavelength[575:], rtol=1e-6,
                                       err_msg=f'ScanNo: {scanno}', verbose=True)
            np.testing.assert_allclose(unapplied.fwhm[575:], applied.fwhm[575:], rtol=1e-6,
                                       err_msg=f'ScanNo: {scanno}', verbose=True)
            # The wavelengths for pixels 574 and down should differ by a constant amount that is not zero
            wave_diff = unapplied.wavelength[:575] - applied.wavelength[:575]
            self.assertNotEqual(0.0, wave_diff[0])
            np.testing.assert_allclose(wave_diff[0], wave_diff, rtol=1e-6, err_msg=f'ScanNo: {scanno}', verbose=True)
            # The fwhm for pixels 574 and down should be constant after the psf is applied.
            np.testing.assert_allclose(applied.fwhm[0], applied.fwhm[:575], rtol=1e-6,
                                       err_msg=f'ScanNo: {scanno}', verbose=True)

    def test_spectrograph_rsas_shift(self):
        """ Test that the RSAS shift is applied """
        for scanno in self.scannos:
            unshifted = open_level1_spectrograph(scanno=scanno)
            shifted = open_level1_spectrograph(scanno=scanno, apply_rsas=True)

            alt_diff = unshifted.l1.altitude - shifted.l1.altitude

            # Difference between tangent altitudes should be equal to the rsas_shift
            np.testing.assert_allclose(alt_diff, shifted.rsas_shift, atol=0.005)
